#ifndef _OPENDRIVE_LANE_H
#define _OPENDRIVE_LANE_H

#ifdef _HAVE_XSD
#include "xml/OpenDRIVE_1.4H.hxx"
#endif

#include "opendrive_common.h"
#include "opendrive_road_node.h"
#include "opendrive_road_segment.h"
#include "opendrive_lane.h"

class COpendriveLane
{
  friend class COpendriveRoadSegment;
  friend class COpendriveRoadNode;
  friend class COpendriveRoad;
  private:
    std::vector<COpendriveRoadNode *> nodes;
    std::vector<COpendriveLane *> next;
    std::vector<COpendriveLane *> prev;
    COpendriveLane *left_lane;
    opendrive_mark_t left_mark;
    COpendriveLane *right_lane;
    opendrive_mark_t right_mark;
    COpendriveRoadSegment *segment;
    double scale_factor;
    double resolution;
    double width;
    double speed;
    double offset;
    int id;
    unsigned int index;
  protected:
    COpendriveLane();
    COpendriveLane(const COpendriveLane &object);
    void load(const ::lane &lane_info,COpendriveRoadSegment *segment);
    void link_neightbour_lane(COpendriveLane *lane);
    void link_lane(COpendriveLane *lane,opendrive_mark_t mark,bool from_start,bool to_start,bool belongs_to_lane);
    void add_next_lane(COpendriveLane *lane);
    void add_prev_lane(COpendriveLane *lane);
    void remove_lane(COpendriveLane *lane);
    void add_node(COpendriveRoadNode *node);
    bool has_node(COpendriveRoadNode *node);
    bool has_node_with_index(unsigned int index);
    void set_parent_segment(COpendriveRoadSegment *segment);
    void set_left_lane(COpendriveLane *lane,opendrive_mark_t mark);
    void set_right_lane(COpendriveLane *lane,opendrive_mark_t mark);
    void set_resolution(double res);
    void set_scale_factor(double scale);
    void set_width(double width);
    void set_speed(double speed);
    void set_offset(double offset);
    void set_id(int id);
    void set_index(unsigned int index);
    bool updated(lane_up_ref_t &refs);
    void update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void clean_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void update_start_node(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref,TOpendriveWorldPose *start=NULL);
    void update_end_node(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref,TOpendriveWorldPose *end=NULL);
    COpendriveLane *get_sub_lane(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref,TOpendriveWorldPose *start=NULL,TOpendriveWorldPose *end=NULL);
    COpendriveLane *clone(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref);
    TOpendriveLocalPose transform_to_local_pose(TOpendriveTrackPose &track) const;
    TOpendriveWorldPose transform_to_world_pose(TOpendriveTrackPose &track) const;
  public:
    /**
     * \brief Returns the number of road nodes in the lane
     *
     * This function returns how many road nodes belong to the lane.
     *
     * \return This function returns the number of road nodes. The range of 
     *         valid indexes goes from 0 to the value returned by this function -1.
     */
    unsigned int get_num_nodes(void) const;
    /**
     * \brief Returns a reference to a road node by local index
     *
     * This function returns a constant reference to the desired road node 
     * object, which can not be modified by the user. If the index is not valid,
     * this function will throw an exception.
     *
     * \param index is the 0 based local index of the desired road node. It is not
     *              the unique index assigned to the road nodes when they are created.
     *
     * \return constant reference to the desired road node object.
     */
    const COpendriveRoadNode &get_node(unsigned int index) const;
    /**
     * \brief Returns a reference to the parent road segment
     * 
     * This function returns a constant reference to the road segment where the lane
     * belongs, which can not be modified by the user. If the parent segment is not
     * defined, this function will throw an exception.
     *
     * \return constant reference to the parent road segment. 
     */
    const COpendriveRoadSegment &get_segment(void) const;
    /**
     * \brief Returns the number of next lanes 
     *
     * This function returns how many lanes are connected to this one in the direction
     * of the traffic (the next lanes).
     *
     * \return This function returns the number of next lanes. The range of 
     *         valid indexes goes from 0 to the value returned by this function -1.
     */
    unsigned int get_num_next_lanes(void) const;
    /**
     * \brief Returns a reference to a next lane
     *
     * This function returns a constant reference to a next lane connected to this one,
     * which can not be modified by the user. If the index is not valid, this function 
     * will throw an exception.
     *
     * \param index is the 0 based local index of the desired next lane. It is not
     *              the unique index assigned to the lanes when they are created.
     *
     * \return constant reference to a next lane.
     */
    const COpendriveLane &get_next_lane(unsigned int index) const;
    /**
     * \brief Returns the number of previous lanes 
     *
     * This function returns how many lanes are connected to this one in the opposite 
     * direction of the traffic (the previous lanes).
     *
     * \return This function returns the number of previous lanes. The range of 
     *         valid indexes goes from 0 to the value returned by this function -1.
     */
    unsigned int get_num_prev_lanes(void) const;
    /**
     * \brief Returns a reference to a previous lane
     *
     * This function returns a constant reference to a previous lane connected to this one,
     * which can not be modified by the user. If the index is not valid, this function 
     * will throw an exception.
     *
     * \param index is the 0 based local index of the desired previous lane. It is not
     *              the unique index assigned to the lanes when they are created.
     *
     * \return constant reference to a previous lane.
     */
    const COpendriveLane &get_prev_lane(unsigned int index) const;
    /**
     * \brief Checks whether there exist a left lane or not
     *
     * This function checks if the lane has an adjacent left lane. Adjacent lanes are 
     * on the same side of the road. Lanes of opposite sides of the road are not 
     * considered adjacent.
     *
     * \return true if the lane has an adjacent left lane or false otherwise.
     */
    bool exist_left_lane(void);
    /**
     * \brief Returns a reference to a left lane
     *
     * This function returns a constant reference to a left lane adjacent to this one,
     * which can not be modified by the user. If there is no adjacent left lane, this 
     * function will throw an exception.
     *
     * \return constant reference to the left lane.
     */
    const COpendriveLane &get_left_lane(void) const;
    /**
     * \brief Checks whether there exist a right lane or not
     *
     * This function checks if the lane has an adjacent right lane. Adjacent lanes are 
     * on the same side of the road. Lanes of opposite sides of the road are not 
     * considered adjacent.
     *
     * \return true if the lane has an adjacent right lane or false otherwise.
     */
    bool exist_right_lane(void);
    /**
     * \brief Returns a reference to a right lane
     *
     * This function returns a constant reference to a right lane adjacent to this one,
     * which can not be modified by the user. If there is no adjacent right lane, this 
     * function will throw an exception.
     *
     * \return constant reference to the right lane.
     */
    const COpendriveLane &get_right_lane(void) const;
    /**
     * \brief return the road mark at the left side of the lane
     *
     * This function returns the road mark identifier at the left side of the lane, even
     * if the left lane is not defined.
     *
     * \return the road mark identifier at the left side of the lane.
     */
    opendrive_mark_t get_left_road_mark(void) const;
    /**
     * \brief return the road mark at the right side of the lane
     *
     * This function returns the road mark identifier at the right side of the lane, even
     * if the right lane is not defined.
     *
     * \return the road mark identifier at the right side of the lane.
     */
    opendrive_mark_t get_right_road_mark(void) const;
    /**
     * \brief Returns the width of the lane
     * 
     * \returns the width in meters of the lane.
     */
    double get_width(void) const;
    /**
     * \brief Returns the speed limit of the lane
     *
     * \return the speed limit of the lane in km/h
     */
    double get_speed(void) const;
    /**
     * \brief Returns the offset from the road segment center.
     *
     * This function returns the offset of the center of the lane with respect of
     * the center of the road segment where it belongs.
     *
     * \return the distance form the center of the segment in meters.
     */
    double get_center_offset(void) const;
    /**
     * \brief Returns the unique index of the lane
     *
     * \return the unique index assigned to the lane when it was created.
     */
    unsigned int get_index(void) const;
    /**
     * \brief Returns the opendrive identifier of the lane
     *
     * This function returns the opendrive identifier of the lane: negative integers
     * values are used to identify right lanes, and positive values to identify left 
     * lanes. These values increase (for left lanes) or decrease (for right lanes) the 
     * further they are from the center lane of the segment.
     *
     * \return the opendrive identifier of the lane.
     */
    int get_id(void) const;
    /**
     * \brief Returns the start pose of the lane
     *
     * This function returns the start pose (x and y coordinates and heading) of the 
     * lane in the world reference system. The lanes are defined in the direction of 
     * the traffic. That is, right lanes have the same heading as the road segment, 
     * but left lanes have opposite heading, starting at the end of the road segment 
     * and ending at the start.
     *
     * \return the start pose of the lane in world coordinates.
     */
    TOpendriveWorldPose get_start_pose(void) const;
    /**
     * \brief Returns the end pose of the lane
     *
     * This function returns the end pose (x and y coordinates and heading) of the 
     * lane in the world reference system. The lanes are defined in the direction of 
     * the traffic. That is, right lanes have the same heading as the road segment, 
     * but left lanes have opposite heading, starting at the end of the road segment 
     * and ending at the start.
     *
     * \return the end pose of the lane in world coordinates.
     */
    TOpendriveWorldPose get_end_pose(void) const;
    /** 
     * \brief returns the length of the lane
     *
     * This function returns the length of the lane geometry. The lane geometry is 
     * the same as the link geometries of the lane links connecting the road nodes
     * belonging to the lane.
     *
     * \return length of the lane in meters.
     */
    double get_length(void) const;
    /** 
     * \brief returns the pose of the lane at a given point
     *
     * This function returns the pose (x and y coordinates and heading) of the lane geometry
     * at a given distance from the origin point of the lane. The lane geometry is 
     * the same as the link geometries of the lane links connecting the road nodes
     * belonging to the lane.
     *
     * \param length desired distance from the origin point of the lane in meters. This 
     *               distance must be smaller or equal than the actual lane length.
     *
     * \return the pose at the desired point in the world reference system.
     */
    TOpendriveWorldPose get_pose_at(double length) const;
    /** 
     * \brief returns the curvature of the lane at a given point
     *
     * This function returns the curvature of the lane geometry at a given distance from 
     * the origin point of the lane. The lane geometry is the same as the link geometries 
     * of the lane links connecting the road nodes belonging to the lane.
     *
     * \param length desired distance from the origin point of the lane in meters. This 
     *               distance must be smaller or equal than the actual lane length.
     *
     * \return the curvature at the desired point.
     */
    double get_curvature_at(double length) const;    
    /** 
     * \brief returns the closest node to a given pose
     *
     * This functions finds the closest road node belonging to the lane to a given 
     * pose, and returns it. This function only returns a valid reference when the 
     * given pose coincides with the lane. Otherwise, it throws an exception.
     *
     * \param world given pose to find the closest road node.
     * \param distance the node distance between the projection of the given pose to 
     *                 the lane link geometry of the closest road node, and the start 
     *                 position of the node, following the corresponding lane link geometry. 
     *                 If the closest road node does not exist this argument is set to the 
     *                 maximum double value (std::numeric_limits<double>::max()) 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane link geometry of the closest road node
     *  
     * \return a constant reference to the closest node which can not be modified by the 
     *         user.
     */
    const COpendriveRoadNode &get_closest_node(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /** 
     * \brief returns the index of the closest node to a given pose
     *
     * This functions finds the closest road node belonging to the lane to a given 
     * pose, and returns its index. This function only returns a valid index 
     * when the given pose coincides with the lane.
     *
     * \param world given pose to find the closest road node.
     * \param distance the node distance between the projection of the given pose to 
     *                 the lane link geometry of the closest road node, and the start 
     *                 position of the node, following the corresponding lane link geometry. 
     *                 If the closest road node does not exist this argument is set to the 
     *                 maximum double value (std::numeric_limits<double>::max()) 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane link geometry of the closest road node
     *  
     * \return the index of the closest node or -1 if there is no node close to the given 
     *         pose.
     */
    unsigned int get_closest_node_index(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /**
     * \brief returns the closest pose to a given pose
     *
     * This functions finds the closest pose (minimum distance) on the lane geometry 
     * to a given pose. The closest pose only exist if the imaginary line between the 
     * given pose and its projection into the lane is orthogonal to the lane geometry. 
     * Otherwise, it is considered that the given pose does not coincide with the lane.
     * 
     * \param world given pose to find the closest pose.
     * \param closest_pose the closest pose to the lane if the given pose coincides with 
     *                     it. Otherwise the data on this parameters is not valid.
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane geometry.
     *
     * \return the lane distance between the projection of the given pose to the 
     *         lane geometry, following its geometry. If the closest pose does not exist,
     *         the maximum double value (std::numeric_limits<double>::max()) will be 
     *         returned.
     */
    double get_closest_pose(TOpendriveWorldPose &pose,TOpendriveWorldPose &closest_pose,double angle_tol=0.1) const;
    /** 
     * \brief overloaded stream operator
     * 
     * This functions streams human readable information about the lane. This information
     * includes:
     *  * The opendrive lane id.
     *  * Its width, sped limit and offset from the center of the road segment.
     *  * The parent road segment.
     *  * For each previous and next lanes: the road segment and lane information.
     *  * The left lane and the left road mark.
     *  * The right lane and the right road mark.
     *  * The road nodes information.
     *
     */
    friend std::ostream& operator<<(std::ostream& out, COpendriveLane &lane);
    /**
     * brief Destructor
     */
    ~COpendriveLane();
};

#endif
