#ifndef _OPENDRIVE_LINK_H
#define _OPENDRIVE_LINK_H

#include "opendrive_common.h"
#include "g2_spline.h"
#include "opendrive_line.h"
#include "opendrive_spiral.h"
#include "opendrive_arc.h"
#include "opendrive_param_poly3.h"
#include "opendrive_road_node.h"

class COpendriveLink
{
  friend class COpendriveRoadSegment;
  friend class COpendriveRoadNode;
  friend class COpendriveRoad;
  private:
    COpendriveRoadNode *prev;
    COpendriveRoadNode *next;
    COpendriveRoadSegment *segment;
    COpendriveLane *lane;
    CG2Spline *spline;
    opendrive_mark_t mark;
    double resolution;
    double scale_factor;    
  protected:
    COpendriveLink();
    COpendriveLink(const COpendriveLink &object);
    void set_prev(COpendriveRoadNode *node);
    void set_next(COpendriveRoadNode *node);
    void set_road_mark(opendrive_mark_t mark);
    void set_parent_segment(COpendriveRoadSegment *segment);
    void set_parent_lane(COpendriveLane *lane);
    void set_resolution(double res);
    void set_scale_factor(double scale);
    void generate(double start_curvature,double end_curvature);
    void update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    bool clean_references(node_up_ref_t &refs);
    void update_start_pose(TOpendriveWorldPose &start,double curvature);
    void update_end_pose(TOpendriveWorldPose &end,double curvature);
  public:
    /** 
     * \brief Returns a reference to the previous road node
     *
     * This function returns a constant reference to the previous COpendriveRoadNode 
     * object. If the previous road node is not defines, an exception will be thrown.
     *
     * \return a constant reference to the previous COpendriveRoadNode object.
     */
    const COpendriveRoadNode &get_prev(void) const;
    /** 
     * \brief Returns a reference to the next road node
     *
     * This function returns a constant reference to the next COpendriveRoadNode 
     * object. If the next road node is not defines, an exception will be thrown.
     *
     * \return a constant reference to the next COpendriveRoadNode object.
     */
    const COpendriveRoadNode &get_next(void) const;
    /** 
     * \brief Returns the road mark identifier between the two road nodes
     *
     * This function returns the possible road mark between the two COpendriveRoadNode 
     * objects linked by each instance of this class. For lane links this will always 
     * be OD_MARK_NONE, but for non-lane links it will correspond to the actual road 
     * mark.
     *
     * \return the road mark identifier.
     */
    opendrive_mark_t get_road_mark(void) const;
    /** 
     * \brief Returns the reference to the parent road segment
     *
     * This function returns a constant reference to the COpendriveRoadSegment object
     * where the each instance of this class belongs. If the reference is not valid,
     * an exception will be thrown.
     *
     * \return a constant reference to the parent COpendriveRoadSegment object.
     */
    const COpendriveRoadSegment &get_parent_segment(void) const;
    /** 
     * \brief Returns the reference to the parent lane
     *
     * This function returns a constant reference to the COpendriveLane object
     * where the each instance of this class belongs. This function will fail for 
     * non-lane links since they do not have a parent lane. In this case an 
     * exception will be thrown.
     * 
     * Use the is_lane_link() function to check the type of link.
     *
     * \return a constant reference to the parent COpendriveRoadSegment object.
     */
    const COpendriveLane &get_parent_lane(void) const;
    /** 
     * \brief returns the type of link
     *
     * This function returns whether each instance of this class is a lane link 
     * or not.
     *
     * \return true if the instance is a lane link and false otherwise.
     */
    bool is_lane_link(void) const;
    /**
     * \brief Returns the resolution
     *
     * \return This function returns the current resolution
     */
    double get_resolution(void) const;
    /**
     * \brief Returns the scale factor
     *
     * \return This function returns the current scale factor
     */
    double get_scale_factor(void) const;
    /** 
     * \brief returns the closest pose to a given one
     *
     * This functions finds the closest pose (minimum distance) on the link geometry 
     * to a given position. The closest pose only exist if the imaginary line between the 
     * given pose and its projection into the link is orthogonal to the link geometry.
     * Otherwise, it is considered that the given pose does not coincide with the link
     * object.
     *
     * \param world given pose to find the closest point. The heading information is
     *              ignored.
     * \param pose the closest pose on the link geometry if the given pose coincide 
     *             with the link object. Otherwise the data on this parameters is not
     *             valid.
     *  
     * \return the link distance between the projected pose and the link start position, 
     *         following the link geometry, if given pose coincide with the link object. 
     *         Otherwise the maximum double value (std::numeric_limits<double>::max()) 
     *         will be returned.
     */
    double find_closest_world_pose(TOpendriveWorldPose &world,TOpendriveWorldPose &pose) const;
    /** 
     * \brief returns the link trajectory (only position)
     * 
     * This function returns the x and y coordinates of the link geometry evaluated at
     * the internal resolution. By default the whole trajectory is returned, but is 
     * possible to get only a sub-trajectory by specifying the start and end lengths.
     *
     * \param x a vector with the x coordinates of the link trajectory (sub-trajectory)
     *          in the world reference system.
     * \param y a vector with the y coordinates of the link trajectory (sub-trajectory)
     *          in the world reference system.
     * \param start_length initial distance from the origin position of the link in 
     *                     meters. By default this value is 0.0.
     * \param end_length final distance from the origin position of the link in meters. 
     *                   By default this value is -1.0 to indicate the end of the link.
     */
    void get_trajectory(std::vector<double> &x, std::vector<double> &y,double start_length=0.0, double end_length=-1.0) const;
    /** 
     * \brief returns the link trajectory (position and heading)
     *
     * This function returns the x and y coordinates, as well as the heading of the 
     * link geometry evaluated at the internal resolution. By default the whole 
     * trajectory is returned, but is possible to get only a sub-trajectory by 
     * specifying the start and end lengths.
     *
     * \param x a vector with the x coordinates of the link trajectory (sub-trajectory)
     *          in the world reference system.
     * \param y a vector with the y coordinates of the link trajectory (sub-trajectory)
     *          in the world reference system.
     * \param yaw a vector with the heading of the link trajectory (sub-trajectory)
     *          in the world reference system.
     * \param start_length initial distance from the origin position of the link in 
     *                     meters. By default this value is 0.0.
     * \param end_length final distance from the origin position of the link in meters. 
     *                   By default this value is -1.0 to indicate the end of the link.
     */
    void get_trajectory(std::vector<double> &x, std::vector<double> &y, std::vector<double> &yaw,double start_length=0.0, double end_length=-1.0) const;
    /** 
     * \brief returns the length of the link
     *
     * This function returns the length of the link geometry.
     *
     * \return length of the link geometry in meters.
     */
    double get_length(void) const;
    /** 
     * \brief returns the pose of the link at a given point
     *
     * This function returns the pose (x and y coordinates and heading) of the link geometry
     * at a given distance from the origin point of the link.
     *
     * \param length desired distance from the origin point of the link in meters. This 
     *               distance must be smaller or equal than the actual link length.
     *
     * \return the pose at the desired point in the world reference system.
     */
    TOpendriveWorldPose get_pose_at(double length);
    /** 
     * \brief returns the curvature of the link at a given point
     *
     * This function returns the curvature of the link geometry at a given distance from 
     * the origin point of the link.
     *
     * \param length desired distance from the origin point of the link in meters. This 
     *               distance must be smaller or equal than the actual link length.
     *
     * \return the curvature at the desired point.
     */
    double get_curvature_at(double length);
    /** 
     * \brief overloaded stream operator
     * 
     * This functions streams human readable information about the link. This information
     * includes:
     *  * Index of the previous and next COpendriveRoad node objects.
     *  * Name of the parent COpendriveROadSegment object.
     *  * Id of the parent lane for lane links.
     *  * The road mark between the two road nodes.
     *
     */
    friend std::ostream& operator<<(std::ostream& out, COpendriveLink &link);
    /** 
     * \brief Destructor
     */
    ~COpendriveLink();
};

#endif
