#ifndef _OPENDRIVE_ROAD_H
#define _OPENDRIVE_ROAD_H

#ifdef _HAVE_XSD
#include "xml/OpenDRIVE_1.4H.hxx"
#endif

#include "opendrive_common.h"
#include "opendrive_road_segment.h"
#include "opendrive_road_node.h"
#include "opendrive_lane.h"

class COpendriveRoad
{
  friend class COpendriveRoadSegment;
  friend class COpendriveRoadNode;
  friend class COpendriveLane;
  private:
    std::vector<COpendriveRoadSegment *> segments;
    std::vector<COpendriveRoadNode *> nodes;
    std::vector<COpendriveLane *> lanes;
    double resolution;
    double scale_factor;
    double min_road_length;
  protected:
    void free(void);
    void link_segments(OpenDRIVE &open_drive);
    unsigned int add_node(COpendriveRoadNode *node);
    bool remove_node(COpendriveRoadNode *node);
    COpendriveRoadNode* get_node_at(TOpendriveWorldPose &pose);
    bool node_exists_at(TOpendriveWorldPose &pose);
    unsigned int add_lane(COpendriveLane *lane);
    bool remove_lane(COpendriveLane *lane);
    void complete_open_lanes(void);
    void add_segment(COpendriveRoadSegment *segment);
    bool has_segment(COpendriveRoadSegment *segment);
    std::vector<unsigned int> update_path(node_up_ref_t &node_refs,std::vector<unsigned int> &path);
    void update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void clean_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void reindex(void);
    void prune(std::vector<unsigned int> &path_nodes);
  public:
    /**
     * \brief Default constructor
     */ 
    COpendriveRoad();
    /**
     * \brief Copy constructor
     *
     * \param object constant reference to the object to be copied
     */ 
    COpendriveRoad(const COpendriveRoad& object);
    /**
     * \brief Load an Opendrive road description
     *
     * This function loads the Opendrive description file initializes the internal 
     * data structures. If the file referenced by the parameter does not exist or 
     * the format is not valid, this function will throw an exception.
     *
     * \param filename the name of the opendrive file to load with the extension 
     *                 and the full path.
     */
    void load(const std::string &filename);
    /**
     * \brief Sets the resolution
     *
     * This functions sets the default resolution used to generate all the internal 
     * geometry. This functions automatically changes the resolution of all underlying
     * objects.
     *
     * \param res desired value of the resolution in meters. The default resolution
     *            is 0.1 m.
     */
    void set_resolution(double res);
    /**
     * \brief Sets the scale factor
     *
     * This function sets the scale factor to be applied to all the internal geometry.
     * The scale factor makes it possible to change the size of the road without actually
     * modifying the internal geometry data. This functions automatically changes 
     * the scale factor of all underlying objects.
     *
     * \param scale desired scale factor. A value bigger than 1.0 will scale down the 
     *              road and a value smaller than 1.0 will scale it up. The default 
     *              value is 1.0.
     */
    void set_scale_factor(double scale);
    /**
     * \brief Sets the minim road length
     *
     * This function sets the length threshold to include any imported road segments into 
     * the internal structure. This functions automatically changes the minimum road length
     * of all underlying objects.
     *
     * \param length threshold value to include a road segment. The default value is 0.01.
     */
    void set_min_road_length(double length);
    /**
     * \brief Returns the resolution
     *
     * \return This function returns the current resolution
     */
    double get_resolution(void) const;
    /**
     * \brief Returns the scale factor
     *
     * \return This function returns the current scale factor
     */
    double get_scale_factor(void) const;
    /**
     * \brief Returns the minimum road length
     *
     * \return This function returns the current minimum road length
     */
    double get_min_road_length(void) const;
    /**
     * \brief Returns the number of road segment objects in the whole road
     *
     * \return This function returns the number of road segments present in 
     *         whole road. The range of valid indexes goes from 0 to the 
     *         value returned by this function -1.
     */
    unsigned int get_num_segments(void) const;
    /**
     * \brief Returns a road segment by index
     *
     * This function returns a constant reference to the desired road segment 
     * object, which can not be modified by the user. If the index is not valid,
     * this function will throw an exception.
     *
     * \param index is the 0 based index of the desired road segment
     *
     * \return constant reference to the desired road segment object.
     */
    const COpendriveRoadSegment& get_segment(unsigned int index) const;
    /**
     * \brief Returns a road segment by id
     *
     * This function returns a constant reference to the desired road segment 
     * object, which can not be modified by the user. If the id is not valid,
     * this function will throw an exception.
     *
     * \param id is the Opendrive numeric id of the road segment.
     *
     * \return constant reference to the desired road segment object.
     */
    const COpendriveRoadSegment &get_segment_by_id(unsigned int id) const;
    /**
     * \brief Returns a road segment by name
     *
     * This function returns a constant reference to the desired road segment 
     * object, which can not be modified by the user. If the name is not valid,
     * this function will throw an exception.
     *
     * \param name is the Opendrive name of the road segment.
     *
     * \return constant reference to the desired road segment object.
     */
    const COpendriveRoadSegment &get_segment_by_name(std::string &name) const;
    /**
     * \brief Returns the number of lane objects in the whole road
     *
     * \return This function returns the number of lanes present in whole
     *         road. The range of valid indexes goes from 0 to the value 
     *         returned by this function -1.
     */
    unsigned int get_num_lanes(void) const;
    /**
     * \brief Returns a lane by index
     *
     * This function returns a constant reference to the desired lane 
     * object, which can not be modified by the user. If the index is not valid,
     * this function will throw an exception.
     *
     * \param index is the 0 based index of the desired lane
     *
     * \return constant reference to the desired lane object.
     */
    const COpendriveLane &get_lane(unsigned int index) const;
    /**
     * \brief Returns the number of road node objects in the whole road
     *
     * \return This function returns the number of road nodes present in whole
     *         road. The range of valid indexes goes from 0 to the value 
     *         returned by this function -1.
     */
    unsigned int get_num_nodes(void) const;
    /**
     * \brief Returns a road node by index
     *
     * This function returns a constant reference to the desired road node 
     * object, which can not be modified by the user. If the index is not valid,
     * this function will throw an exception.
     *
     * \param index is the 0 based index of the desired road node
     *
     * \return constant reference to the desired road node object.
     */
    const COpendriveRoadNode& get_node(unsigned int index) const;
    /**
     * \brief overloaded index operator
     *
     * This operator is similar to the get_segment() function. It returns a 
     * constant reference to the desired road segment object, which can not 
     * be modified by the user. If the index is not valid,this function will 
     * throw an exception.
     *
     * \param index is the 0 based index of the desired road segment
     *
     * \return constant reference to the desired road segment object.
     */
    const COpendriveRoadSegment &operator[](std::size_t index);
    /** 
     * \brief returns the closest node to a given pose
     *
     * This functions finds the closest road node belonging to the whole road to a
     * given pose, and returns it.  
     *
     * \param world given pose to find the closest road node.
     * \param distance the node distance between the projection of the given pose to 
     *                 the lane link geometry of the closest road node, and the start 
     *                 position of the node, following the corresponding lane link geometry. 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane link geometry of the closest road node.
     *  
     * \return a constant reference to the closest node which can not be modified by the 
     *         user.
     */
    const COpendriveRoadNode &get_closest_node(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /** 
     * \brief returns the index of the closest node to a given pose
     *
     * This functions finds the closest road node belonging to the whole road to a given 
     * pose, and returns its index.  
     *
     * \param world given pose to find the closest road node.
     * \param distance the node distance between the projection of the given pose to 
     *                 the lane link geometry of the closest road node, and the start 
     *                 position of the node, following the corresponding lane link geometry. 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane link geometry of the closest road node.
     *  
     * \return the index of the closest road node.
     */
    unsigned int get_closest_node_index(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /** 
     * \brief returns the closest lane to a given pose
     *
     * This functions finds the closest lane belonging to the whole road to a
     * given pose, and returns it. 
     *
     * \param world given pose to find the closest lane.
     * \param distance the lane distance between the projection of the given pose to 
     *                 the lane geometry, and the start position of the lane, following 
     *                 its geometry. 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane geometry.
     *  
     * \return a constant reference to the closest lane which can not be modified by the 
     *         user.
     */
    const COpendriveLane &get_closest_lane(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /** 
     * \brief returns the index of the closest lane to a given pose
     *
     * This functions finds the closest lane belonging to the whole road to a given 
     * pose, and returns its index. 
     *
     * \param world given pose to find the closest lane.
     * \param distance the lane distance between the projection of the given pose to 
     *                 the lane geometry, and the start position of the lane, following 
     *                 its geometry. 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane geometry.
     *  
     * \return the index of the closest lane.
     */
    unsigned int get_closest_lane_index(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /** 
     * \brief returns the closest road segment to a given pose
     *
     * This functions finds the closest road segment belonging to the whole road to a
     * given pose, and returns it. The geometry used to find the closest road segment is the
     * one that describes the center lane of the segment, not the lane/link geometries.
     *
     * \param world given pose to find the closest road segment.
     * \param distance the segment distance between the projection of the given pose to 
     *                 the road segment geometry, and the start position of the segment, following 
     *                 its geometry. 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the road segment geometry.
     *  
     * \return a constant reference to the closest road segment which can not be modified by the 
     *         user.
     */
    const COpendriveRoadSegment &get_closest_segment(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /** 
     * \brief returns the index of the closest road segment to a given pose
     *
     * This functions finds the closest road segment belonging to the whole road to a given 
     * pose, and returns its index. The geometry used to find the closest road segment is the
     * one that describes the center lane of the segment, not the lane/link geometries.
     *
     * \param world given pose to find the closest road segment.
     * \param distance the segment distance between the projection of the given pose to 
     *                 the road segment geometry, and the start position of the segment, following 
     *                 its geometry. 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the road segment geometry.
     *  
     * \return the index of the closest road segment.
     */
    unsigned int get_closest_segment_index(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /**
     * \brief returns the closest pose 
     *
     * This functions finds the closest pose (minimum distance) to the whole road 
     * to a given pose. 
     * 
     * \param world given pose to find the closest pose.
     * \param closest_pose the closest pose to the closest to any road geometry.
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on any road geometry.
     *
     * \return the link distance between the projection of the given pose to the 
     *         closest road geometry, following its geometry. 
     */
    double get_closest_pose(TOpendriveWorldPose &pose,TOpendriveWorldPose &closest_pose,double angle_tol=0.1) const;
    /**
     * \brief Returns a set of poses on the whole road close to a given pose
     *
     * This function is similar to the get_closest_pose(), but in this case, the closest 
     * pose of any road geometry which is closer to the given pose that a given distance will be 
     * returned.
     * 
     * \param world given pose to find the closest poses.
     * \param closest_poses the closest poses on any road geometry.
     * \param distances the link distance between the projection of the given pose to each  
     *                  of the closest road geometries, following the corresponding geometry.
     * \param dist_tol maximum allowed distance from the given position to any road geometry. 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on any road geometry.
     */
    void get_closest_poses(TOpendriveWorldPose &pose,std::vector<TOpendriveWorldPose> &closest_poses,std::vector<double> &distances,double dist_tol,double angle_tol=0.1) const;
    /**
     * \brief creates a sub-road with only the road elements of a given path
     *  
     * This functions creates a new sub road with only the whole road segments that 
     * include the road nodes of the given path. These new road segments will only have 
     * the lanes and the road nodes on the same side as the given path. 
     *
     * The road elements of the new road keep the required original information and connectivity
     * but are completely independent from the original ones. The only information that 
     * changes in the new sub-road is:
     *
     * * the remaining road segments, lanes and road nodes are re-indexed, so that their 
     *   unique index is coherent with the new data structure. 
     * * for any open road segment (that is, a road segment that is not connected), new road 
     *   nodes are placed at the end of the open road segment to keep its lane and link 
     *   geometric information.
     *
     * If the given path has non-consecutive road nodes, this function will fail throwing an
     * exception.
     *
     * \param path_nodes sequence of road nodes describing the path. The road nodes are 
     *                   identified by their unique index.
     * \param end_pose end pose of the path. When final road node has multiple road segment
     *                 parents, the end pose indicates which one of them must be included 
     *                 in the new road.
     * \param new_road The new sub-road object.
     *
     * \return the sequence of road nodes describing the path in the new sub-road.
     */
    std::vector<unsigned int> get_sub_road(std::vector<unsigned int> &path_nodes,TOpendriveWorldPose &end_pose,COpendriveRoad &new_road);
    /**
     * \brief creates a sub-road with only the road elements within a distance to a given pose
     *
     * This functions creates a new sub road stating a given distance (margin) from the start 
     * pose and reaching up to a given distance following the traffic direction of the start
     * pose. The new road segments will keep all the original lanes and road nodes. 
     *
     * The distance is measured following the road segment geometries, not the euclidean
     * distance. The new road may be shorter than expected if the original road is not long 
     * enough.
     *
     * The road elements of the new road keep the required original information and connectivity
     * but are completely independent from the original ones. The only information that 
     * changes in the new sub-road is:
     *
     * * the remaining road segments, lanes and road nodes are re-indexed, so that their 
     *   unique index is coherent with the new data structure. 
     * * for any open road segment (that is, a road segment that is not connected), new road 
     *   nodes are placed at the end of the open road segment to keep its lane and link 
     *   geometric information.
     *
     * If the part of the road of interest has multiple paths, this function will fail throwing an
     * exception.
     *
     * \param start_pose start pose of the new road. This pose implicitly defines the side
     *                   of the road taken to generate the sub-road.
     * \param length this is the desired length of the new road in meters. On success, this
     *               parameter will be updated with the actual length of the new road.
     * \param margin a distance to be added at the beginning and the end of the new road, if
     *               possible. This makes the actual length of the new road equal to length+
     *               2.0*margin.
     * \param new_road The new sub-road object.
     * 
     */
    void get_sub_road(TOpendriveWorldPose &start_pose,double &length,double margin,COpendriveRoad &new_road);
    /**
     * \brief creates a full copy of the road
     */
    void operator=(const COpendriveRoad& object);
    /** 
     * \brief overloaded stream operator
     * 
     * This functions streams human readable information about the whole road. This information
     * includes:
     *  * The resolution, scale_factor and minimum road length
     *  * The road segments information.
     *
     */
    friend std::ostream& operator<<(std::ostream& out, COpendriveRoad &road);
    /**
     * \brief Destructor
     */
    ~COpendriveRoad();
};

#endif
