#ifndef _OPENDRIVE_ROAD_SEGMENT_H
#define _OPENDRIVE_ROAD_SEGMENT_H

#ifdef _HAVE_XSD
#include "xml/OpenDRIVE_1.4H.hxx"
#endif

#include "opendrive_common.h"
#include "opendrive_lane.h"
#include "opendrive_signal.h"
#include "opendrive_object.h"
#include "opendrive_road.h"
#include "opendrive_road_node.h"
#include "opendrive_geometry.h"

typedef struct{
  COpendriveGeometry *opendrive;
  CG2Spline *spline;
}TOpendriveRoadSegmentGeometry;

class COpendriveRoadSegment
{
  friend class COpendriveRoad;
  friend class COpendriveRoadNode;
  friend class COpendriveLane;
  friend class COpendriveSignal;
  friend class COpendriveObject;
  private:
    std::map<int,COpendriveLane *> lanes;
    std::vector<COpendriveRoadNode *> nodes;
    std::vector<TOpendriveRoadSegmentGeometry> geometries;
    COpendriveRoad *parent_road;
    double resolution;
    double scale_factor;
    double min_road_length;
    int num_right_lanes;
    int num_left_lanes;
    std::vector<COpendriveSignal *> signals;
    std::vector<COpendriveObject *> objects;
    std::vector<COpendriveRoadSegment *> connecting;
    std::string name;
    unsigned int id;
    opendrive_mark_t center_mark;
  protected:
    COpendriveRoadSegment();
    COpendriveRoadSegment(const COpendriveRoadSegment &object);
    void load(OpenDRIVE::road_type &road_info);
    void free(void);
    void set_resolution(double res);
    void set_scale_factor(double scale);
    void set_min_road_length(double length);
    void set_parent_road(COpendriveRoad *parent);
    void set_name(std::string &name);
    void set_id(unsigned int id);
    void set_center_mark(opendrive_mark_t mark);
    bool updated(segment_up_ref_t &segment_refs);
    COpendriveRoadSegment *get_original_segment(segment_up_ref_t &segment_refs);
    void update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void clean_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void add_lanes(lanes::laneSection_type &lane_section);
    void add_lane(const ::lane &lane_info);
    void remove_lane(COpendriveLane *lane);
    void add_geometries(OpenDRIVE::road_type &road_info);
    void add_nodes(void);
    void add_node(TOpendriveRoadSegmentGeometry &geometry,COpendriveLane *lane);
    void add_empty_node(TOpendriveWorldPose &pose,COpendriveLane *lane);
    void remove_node(COpendriveRoadNode *node);
    bool has_node(COpendriveRoadNode *node) const;
    void link_neightbour_lanes(lanes::laneSection_type &lane_section);
    void link_neightbour_lanes(void);
    void link_segment(COpendriveRoadSegment *segment);
    void link_segment(COpendriveRoadSegment *segment,int from,bool from_start, int to,bool to_start);
    bool connects_to(COpendriveRoadSegment *segment);
    bool connects_segments(COpendriveRoadSegment *segment1,COpendriveRoadSegment *segment2);
    COpendriveRoadSegment *get_sub_segment(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref,int node_side,TOpendriveWorldPose *start=NULL,TOpendriveWorldPose *end=NULL) const;
    COpendriveRoadSegment *clone(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref) const;
    TOpendriveLocalPose transform_to_local_pose(const TOpendriveTrackPose &track);
    TOpendriveWorldPose transform_to_world_pose(const TOpendriveTrackPose &track);
  public:
    /**
     * \brief Returns the Opendrive name
     *
     * \return The Opendrive name of the road segment 
     */
    std::string get_name(void) const;
    /**
     * \brief Returns the Opendirve numeric identifier
     *
     * \return The Opendrive numeric identifier of the road segment.
     */
    unsigned int get_id(void) const;
    /**
     * \brief Returns the number of right lanes 
     *
     * This function returns how many lanes there are on the right side of the road
     * segment. The number of left and right lanes can be different.
     *
     * \return This function returns the number of right lanes. The range of 
     *         valid indexes goes from -1 to the value returned by this function 
     *         negated.
     */
    unsigned int get_num_right_lanes(void) const;
    /**
     * \brief Returns the number of left lanes 
     *
     * This function returns how many lanes there are on the left side of the road
     * segment. The number of left and right lanes can be different.
     *
     * \return This function returns the number of left lanes. The range of 
     *         valid indexes goes from 1 to the value returned by this function.
     */
    unsigned int get_num_left_lanes(void) const;
    /**
     * \brief Returns a reference to a lane by id
     *
     * This function returns a constant reference to a lane of the road segment,
     * which can not be modified by the user. If the identifier is not valid, this 
     * function will throw an exception.
     *
     * \param id is the Opendrive Lane identifier: negative integers values are used 
     *           for right lanes, and positive values for left lanes. These values 
     *           increase (for left lanes) or decrease (for right lanes) the further 
     *           they are from the center lane of the segment.
     *
     * \return constant reference to a lane.
     */
    const COpendriveLane &get_lane(int id) const;
    /**
     * \brief Returns the number of road nodes
     *
     * This function returns how many road nodes belong to the road segment.
     *
     * \return This function returns the number of road nodes. The range of 
     *         valid indexes goes from 0 to the value returned by this function -1.
     */
    unsigned int get_num_nodes(void) const;
    /**
     * \brief Returns a reference to a road node by local index
     *
     * This function returns a constant reference to the desired road node 
     * object, which can not be modified by the user. If the index is not valid,
     * this function will throw an exception.
     *
     * \param index is the 0 based local index of the desired road node. It is not
     *              the unique index assigned to the road nodes when they are created.
     *
     * \return constant reference to the desired road node object.
     */
    const COpendriveRoadNode &get_node(unsigned int index) const;
    /**
     * \brief Returns a reference to the parent road
     *
     * This function returns a constant reference to the parent road object, which 
     * can not be modified by the user. If the internal parent road is invalid, this 
     * function will throw an exception.
     *
     * \return a constant reference to the parent road object.
     */
    const COpendriveRoad &get_parent_road(void) const;
    /**
     * \brief Returns the number of road signals
     *
     * This function returns how many road signals are placed at this road segment.
     *
     * \return This function returns the number of road signals. The range of 
     *         valid indexes goes from 0 to the value returned by this function -1.
     */
    unsigned int get_num_signals(void) const;
    /**
     * \brief Returns a reference to a road signal by local index
     *
     * This function returns a constant reference to the desired road signal 
     * object, which can not be modified by the user. If the index is not valid,
     * this function will throw an exception.
     *
     * \param index is the 0 based local index of the desired road signal. 
     *
     * \return constant reference to the desired road signal object.
     */
    const COpendriveSignal &get_signal(unsigned int index) const;
    /**
     * \brief Returns the number of road objects
     *
     * This function returns how many road objects are placed at this road segment.
     *
     * \return This function returns the number of road objects. The range of 
     *         valid indexes goes from 0 to the value returned by this function -1.
     */
    unsigned int get_num_objects(void) const;
    /**
     * \brief Returns a reference to a road object by local index
     *
     * This function returns a constant reference to the desired road object 
     * object, which can not be modified by the user. If the index is not valid,
     * this function will throw an exception.
     *
     * \param index is the 0 based local index of the desired road object. 
     *
     * \return constant reference to the desired road object object.
     */
    const COpendriveObject &get_object(unsigned int index) const;
    /**
     * \brief Returns the number of connecting road segments
     *
     * This function returns how many road segments are connected to this road
     * segment. Since the road segments have to possible traffic directions, the 
     * concept of next and previous is only defined when the side of the road is 
     * selected.
     *
     * \return This function returns the number of connecting road segments. The 
     *         range of valid indexes goes from 0 to the value returned by this 
     *         function -1.
     */
    unsigned int get_num_connecting(void) const;
    /**
     * \brief Returns a reference to a connecting road segment by local index
     *
     * This function returns a constant reference to the desired connecting road 
     * segment object, which can not be modified by the user. If the index is 
     * not valid, this function will throw an exception.
     *
     * \param index is the 0 based local index of the desired connecting road
     *              segment. 
     *
     * \return constant reference to the desired connecting road segment object.
     */
    const COpendriveRoadSegment &get_connecting(unsigned int index) const;
    /** 
     * \brief returns the length of the road segment
     *
     * This function returns the length of the road segment geometry. The road 
     * segment trajectory is represented by Opendive geometry objects and also
     * by G2 Splines, and it is different from the link and lanes geometries. 
     * Both representations are equivalent.
     *
     * \return length of the road segment in meters.
     */
    double get_length(void) const;
    /** 
     * \brief returns the pose of the road segment at a given point
     *
     * This function returns the pose (x and y coordinates and heading) of the center 
     * lane geometry of the road segment at a given distance from the origin point 
     * of the road segment. 
     *
     * The road segment trajectory is represented by Opendive geometry objects and also
     * by G2 Splines, and it is different from the link and lanes geometries. 
     * Both representations are equivalent.
     *
     * \param length desired distance from the origin point of the road segment in meters.
     *               This distance must be smaller or equal than the actual road segment
     *               length.
     *
     * \return the pose at the desired point in the world reference system.
     */
    TOpendriveWorldPose get_pose_at(double length) const;
    /** 
     * \brief returns the curvature of the road segment at a given point
     *
     * This function returns the curvature of the center geometry of the road segment
     * at a given distance from the origin point of the road segment. 
     *
     * The road segment trajectory is represented by Opendive geometry objects and also
     * by G2 Splines, and it is different from the link and lanes geometries. 
     * Both representations are equivalent.
     *
     * \param length desired distance from the origin point of the road segment in meters. 
     *               This distance must be smaller or equal than the actual road segment 
     *               length.
     *
     * \return the curvature at the desired point.
     */
    double get_curvature_at(double length) const;
    /** 
     * \brief returns the closest node to a given pose
     *
     * This functions finds the closest road node belonging to the road segment to a
     * given pose, and returns it. This function only returns a valid reference when the 
     * given pose coincides with the road segment. Otherwise, it throws an exception.
     *
     * \param world given pose to find the closest road node.
     * \param distance the node distance between the projection of the given pose to 
     *                 the lane link geometry of the closest road node, and the start 
     *                 position of the node, following the corresponding lane link geometry. 
     *                 If the closest road node does not exist this argument is set to the 
     *                 maximum double value (std::numeric_limits<double>::max()) 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane link geometry of the closest road node.
     *  
     * \return a constant reference to the closest node which can not be modified by the 
     *         user.
     */
    const COpendriveRoadNode &get_closest_node(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /** 
     * \brief returns the index of the closest node to a given pose
     *
     * This functions finds the closest road node belonging to the road segment to a given 
     * pose, and returns its index. This function only returns a valid index 
     * when the given pose coincides with the road segment.
     *
     * \param world given pose to find the closest road node.
     * \param distance the node distance between the projection of the given pose to 
     *                 the lane link geometry of the closest road node, and the start 
     *                 position of the node, following the corresponding lane link geometry. 
     *                 If the closest road node does not exist this argument is set to the 
     *                 maximum double value (std::numeric_limits<double>::max()) 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane link geometry of the closest road node.
     *  
     * \return the index of the closest node or -1 if there is no node close to the given 
     *         pose.
     */
    unsigned int get_closest_node_index(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /** 
     * \brief returns the closest lane to a given pose
     *
     * This functions finds the closest lane belonging to the road segment to a
     * given pose, and returns it. This function only returns a valid reference when the 
     * given pose coincides with the road segment. Otherwise, it throws an exception.
     *
     * \param world given pose to find the closest lane.
     * \param distance the lane distance between the projection of the given pose to 
     *                 the lane geometry, and the start position of the lane, following 
     *                 its geometry. If the closest lane does not exist this argument 
     *                 is set to the maximum double value (std::numeric_limits<double>::max()) 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane geometry.
     *  
     * \return a constant reference to the closest lane which can not be modified by the 
     *         user.
     */
    const COpendriveLane &get_closest_lane(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /** 
     * \brief returns the id of the closest lane to a given pose
     *
     * This functions finds the closest lane belonging to the road segment to a given 
     * pose, and returns its id. This function only returns a valid id 
     * when the given pose coincides with the road segment.
     *
     * \param world given pose to find the closest lane.
     * \param distance the lane distance between the projection of the given pose to 
     *                 the lane geometry, and the start position of the lane, following 
     *                 its geometry. If the closest lane does not exist this argument 
     *                 is set to the maximum double value (std::numeric_limits<double>::max()) 
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the lane geometry.
     *  
     * \return the id of the closest node or 0 if there is no lane close to the given 
     *         pose.
     */
    int get_closest_lane_id(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    /**
     * \brief returns the closest pose to a given pose
     *
     * This functions finds the closest pose (minimum distance) on the road segment geometry 
     * to a given pose. The closest pose only exist if the imaginary line between the 
     * given pose and its projection into the road segment is orthogonal to the segment geometry. 
     * Otherwise, it is considered that the given pose does not coincide with the road segment.
     *
     * \param world given pose to find the closest pose.
     * \param closest_pose the closest pose to the center lane of the road segment if the 
     *                     given pose coincides with it. Otherwise the data on this 
     *                     parameters is not valid.
     * \param angle_tol maximum heading difference between the given pose and the closest pose
     *                  on the road segment geometry.
     *
     * \return the road segment distance between the projection of the given pose to the 
     *         center lane  geometry of the road segment, following its geometry. If the 
     *         closest pose does not exist, the maximum double value 
     *         (std::numeric_limits<double>::max()) will be returned.
     */
    double get_closest_pose(TOpendriveWorldPose &pose,TOpendriveWorldPose &closest_pose,double angle_tol=0.1) const;
    /**
     * \brief returns a reference to the next road segments.
     *
     * This function returns a constant reference to the set of road segments that will follow
     * this road segment, depending on the road side of interest. 
     *
     * \param input_side an integer indicating the road side of interest of the current 
     *                   road segment. It is negative for the right side and positive for
     *                   the left one (similar to the lane id criteria).
     * \param output_sides output parameter that indicates the side of the next road segments
     *                     that connects with the side of interest of the current one. It 
     *                     is negative for the right side and positive for the left one. If
     *                     there are no next road segments, this vector will be empty.
     *
     * \return a vector with the constant references of the next road segments. If there 
     *         are no next road segments, this vector will be empty.
     */
    std::vector<const COpendriveRoadSegment *> get_next_segments(int input_side,std::vector<int> &output_sides) const;
    /**
     * \brief returns a reference to the previous road segments.
     *
     * This function returns a constant reference to the set of road segments that preceed
     * this road segment, depending on the road side of interest. 
     *
     * \param input_side an integer indicating the road side of interest of the current 
     *                   road segment. It is negative for the right side and positive for
     *                   the left one (similar to the lane id criteria).
     * \param output_sides output parameter that indicates the side of the previous road segments
     *                     that connects with the side of interest of the current one. It 
     *                     is negative for the right side and positive for the left one. If
     *                     there are no previous road segments, this vector will be empty.
     *
     * \return a vector with the constant references of the previous road segments. If there 
     *         are no previous road segments, this vector will be empty.
     */
    std::vector<const COpendriveRoadSegment *> get_prev_segments(int input_side,std::vector<int> &output_sides) const;
    /**
     * \brief Returns the road segment side of a given pose
     *
     * This function returns the road segment side of a given pose. If the given pose
     * does not coincide with the road segment, an exception will be thrown.
     *
     * \return an integer indicating the road side of the given pose. It is negative for 
     *         the right side and positive for the left one (similar to the lane id criteria).
     */
    int get_pose_side(TOpendriveWorldPose &pose) const;
    /** 
     * \brief overloaded stream operator
     * 
     * This functions streams human readable information about the road segment. This information
     * includes:
     *  * The name of the connecting road segments.
     *  * The road nodes information.
     *  * The left lanes information.
     *  * The right lanes information.
     *  * The road signals information.
     *  * The road objects information.
     *
     */
    friend std::ostream& operator<<(std::ostream& out, COpendriveRoadSegment &segment);
    /**
     * \brief Destructor 
     */
    ~COpendriveRoadSegment();
};

#endif
