#include "opendrive_object.h"
#include "exceptions.h"
#include <cmath>

COpendriveObject::COpendriveObject()
{
  this->segment=NULL;
  this->pose.s=0.0;
  this->pose.t=0.0;
  this->pose.heading=0.0;
  this->type="";
  this->name="";
  this->object_type=OD_NONE;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
}

COpendriveObject::COpendriveObject(const COpendriveObject& object)
{
  this->segment=object.segment;
  this->pose.s=object.pose.s;
  this->pose.s=object.pose.t;
  this->pose.heading=object.pose.heading;
  this->type=object.type;
  this->name=object.name;
  this->object_type=object.object_type;
  switch(this->object_type)
  {
    case OD_BOX: 
      this->object.box.length=object.object.box.length;
      this->object.box.width=object.object.box.width;
      this->object.box.height=object.object.box.height;
      break;
    case OD_CYLINDER:
      this->object.cylinder.radius=object.object.cylinder.radius;
      this->object.cylinder.height=object.object.cylinder.height;
      break;
    case OD_POLYGON:
      for(unsigned int i=0;i<object.object.polygon.num_vertices;i++)
      {
        this->object.polygon.vertices[i]=object.object.polygon.vertices[i];
        this->object.polygon.height[i]=object.object.polygon.height[i];
      }
      this->object.polygon.num_vertices=object.object.polygon.num_vertices;
      break;
    case OD_NONE:
      break;
  }
  this->scale_factor=object.scale_factor;
}

void COpendriveObject::load(objects::object_type &object_info,COpendriveRoadSegment *segment)
{
  double u,v;
  unsigned int i;
  std::stringstream ss(object_info.id().get());
  outline::cornerRoad_iterator corner_road_it;
  outline::cornerLocal_iterator corner_local_it;

  ss >> this->id;
  this->pose.s = (object_info.s().present() ? object_info.s().get() : 0.0);
  this->pose.t = (object_info.t().present() ? object_info.t().get() : 0.0);
  this->pose.heading = (object_info.hdg().present() ? object_info.hdg().get() : 0.0);
  this->type = (object_info.type().present() ? object_info.type().get() : "");
  this->name = (object_info.name().present() ? object_info.name().get() : "");
  if(object_info.length().present() && object_info.width().present() && object_info.height().present())
  {
    this->object_type=OD_BOX;
    this->object.box.length=object_info.length().get();
    this->object.box.width=object_info.width().get();
    this->object.box.height=object_info.height().get();
  }
  else if(object_info.height().present() && object_info.radius().present())
  {
    this->object_type=OD_CYLINDER;
    this->object.cylinder.radius=object_info.radius().get();
    this->object.cylinder.height=object_info.height().get();
  }
  else if(object_info.outline().present())
  {
    this->object_type=OD_POLYGON;
    this->object.polygon.num_vertices=0;
    // absolute track coordinates
    if(object_info.outline().get().cornerRoad().size()>0)// absolute track coordinates
    {
      for(i=0,corner_road_it=object_info.outline().get().cornerRoad().begin();i<OD_MAX_VERTICES && corner_road_it != object_info.outline().get().cornerRoad().end(); ++corner_road_it,++i)
      {
        this->object.polygon.vertices[i].s=(corner_road_it->s().present() ? corner_road_it->s().get() : 0.0);
        this->object.polygon.vertices[i].t=(corner_road_it->t().present() ? corner_road_it->t().get() : 0.0);
        this->object.polygon.height[i]=(corner_road_it->height().present() ? corner_road_it->height().get() : 0.0);
        this->object.polygon.num_vertices++;
      }
    }
    else if(object_info.outline().get().cornerLocal().size()>0)// local coordinates with respect to the object anchor pose
    {
      for(i=0,corner_local_it=object_info.outline().get().cornerLocal().begin();i<OD_MAX_VERTICES && corner_local_it != object_info.outline().get().cornerLocal().end(); ++corner_local_it,++i)
      {
        u=(corner_local_it->u().present() ? corner_local_it->u().get() : 0.0);
        v=(corner_local_it->v().present() ? corner_local_it->v().get() : 0.0);
        this->object.polygon.vertices[i].s=this->pose.s + u*cos(this->pose.heading) - v*sin(this->pose.heading);
        this->object.polygon.vertices[i].t=this->pose.t + u*sin(this->pose.heading) + v*cos(this->pose.heading);
        this->object.polygon.height[i]=(corner_local_it->height().present() ? corner_local_it->height().get() : 0.0);
        this->object.polygon.num_vertices++;
      }
    }
  }
  else
    this->object_type=OD_NONE;
  this->segment=segment;
}

void COpendriveObject::update_references(segment_up_ref_t &segment_refs)
{
  if(segment_refs.find(this->segment)!=segment_refs.end())
    this->segment=segment_refs[this->segment];
}

void COpendriveObject::set_scale_factor(double scale)
{
  this->scale_factor=scale;
}

double COpendriveObject::get_scale_factor(void)
{
  return this->scale_factor;
}

TOpendriveTrackPose COpendriveObject::get_track_pose(void) const
{
  TOpendriveTrackPose track;

  track.s=this->pose.s/this->scale_factor;
  track.t=this->pose.t/this->scale_factor;
  track.heading=this->pose.heading;

  return track;
}

TOpendriveWorldPose COpendriveObject::get_world_pose(void) const
{
  TOpendriveWorldPose world;

  if(this->segment!=NULL)
  {
    world=segment->transform_to_world_pose(this->get_track_pose());
    return world;
  }
  else
    throw CException(_HERE_,"Invalid parent segment reference");
}

int COpendriveObject::get_id(void) const
{
  return this->id;
}

std::string COpendriveObject::get_type(void) const
{
  return this->type;
}

std::string COpendriveObject::get_name(void) const
{
  return this->name;
}

bool COpendriveObject::is_box(void) const
{
  if(this->object_type==OD_BOX)
    return true;
  else
    return false;
}

bool COpendriveObject::is_cylinder(void) const
{
  if(this->object_type==OD_CYLINDER)
    return true;
  else
    return false;
}

bool COpendriveObject::is_polygon(void) const
{
  if(this->object_type==OD_POLYGON)
    return true;
  else
    return false;
}

TOpendriveBox COpendriveObject::get_box_data(void) const
{
  TOpendriveBox data;

  data.length=this->object.box.length/this->scale_factor;
  data.width=this->object.box.width/this->scale_factor;
  data.height=this->object.box.height/this->scale_factor;

  return data;
}

TOpendriveCylinder COpendriveObject::get_cylinder_data(void) const
{
  TOpendriveCylinder data;

  data.radius=this->object.cylinder.radius/this->scale_factor;
  data.height=this->object.cylinder.height/this->scale_factor;

  return data;
}

TOpendrivePolygon COpendriveObject::get_polygon_data(void) const
{
  TOpendrivePolygon data;

  for(unsigned int i=0;i<this->object.polygon.num_vertices;i++)
  {
    data.height[i]=this->object.polygon.height[i]/this->scale_factor;
    data.vertices[i].s=this->object.polygon.vertices[i].s/this->scale_factor;
    data.vertices[i].t=this->object.polygon.vertices[i].t/this->scale_factor;
    data.vertices[i].heading=this->object.polygon.vertices[i].heading;
  }
  return data;
}

std::ostream& operator<<(std::ostream& out, COpendriveObject &object)
{
  TOpendriveBox box_data;
  TOpendriveCylinder cyl_data;
  TOpendrivePolygon pol_data;

  out << "  Object id = " << object.id << std::endl;
  out << "    type = " << object.type << std::endl;
  out << "    name = " << object.name << std::endl;
  out << "    pose: s = " << object.pose.s  << ", t = " << object.pose.t << ", heading = " << object.pose.heading << std::endl;
  switch(object.object_type)
  {
    case OD_BOX:
      box_data=object.get_box_data();
      out << "    box:" << std::endl;
      out << "      length = " << box_data.length << std::endl;
      out << "      width = " << box_data.width << std::endl;
      out << "      height = " << box_data.height << std::endl;
      break;
    case OD_CYLINDER:
      cyl_data=object.get_cylinder_data();
      out << "    cylinder:" << std::endl;
      out << "      radius = " << cyl_data.radius << std::endl;
      out << "      height = " << cyl_data.height << std::endl;
      break;
    case OD_POLYGON:
      pol_data=object.get_polygon_data();
      out << "    polygon: " << pol_data.num_vertices << " vertices" << std::endl;
      for (unsigned int i = 0; i < pol_data.num_vertices; i++)
        out << "      vertex " << i << ": s = " << pol_data.vertices[i].s  << ", t = " << pol_data.vertices[i].t << ", height: " << pol_data.height[i] << std::endl;
      break;
    case OD_NONE:
      break;
  }

  return out;
}

COpendriveObject::~COpendriveObject()
{
  this->pose.s=0.0;
  this->pose.t=0.0;
  this->pose.heading=0.0;
  this->type="";
  this->name="";
  this->object_type=OD_NONE;
}

